import warnings

from django.test import SimpleTestCase

from seal.exceptions import UnsealedAttributeAccess


class UnsealedAttributeAccessTests(SimpleTestCase):
    def test_getattr_error_propagation(self):
        class Foo:
            @property
            def bar(self):
                warnings.warn("Test warning", category=UnsealedAttributeAccess)
                return "bar"

        warnings.filterwarnings("error", category=UnsealedAttributeAccess)
        self.addCleanup(warnings.resetwarnings)
        with self.assertRaisesMessage(UnsealedAttributeAccess, "Test warning"):
            getattr(Foo(), "bar", None)



from django.apps import AppConfig, apps


class SealAppConfig(AppConfig):
    name = __package__

    def ready(self):
        from .descriptors import make_contenttypes_sealable
        from .models import SealableModel, make_model_sealable

        try:
            apps.get_app_config("contenttypes")
        except LookupError:
            pass
        else:
            make_contenttypes_sealable()

        for model in apps.get_models():
            opts = model._meta
            if opts.proxy or not issubclass(model, SealableModel):
                continue
            make_model_sealable(model)



from functools import partial
from operator import attrgetter

from django.db import models
from django.db.models.query_utils import select_related_descend

cached_value_getter = attrgetter("get_cached_value")


def get_restricted_select_related_getters(lookups, opts):
    """Turn a select_related dict structure into a tree of attribute getters"""
    for lookup, nested_lookups in lookups.items():
        field = opts.get_field(lookup)
        lookup_opts = field.related_model._meta
        yield (
            cached_value_getter(field),
            tuple(get_restricted_select_related_getters(nested_lookups, lookup_opts)),
        )


def get_unrestricted_select_related_getters(opts, max_depth, cur_depth=1):
    if cur_depth > max_depth:
        return
    for field in opts.fields:
        if not select_related_descend(field, False, None, {}):
            continue
        related_model_meta = field.related_model._meta
        yield (
            cached_value_getter(field),
            tuple(
                get_unrestricted_select_related_getters(
                    related_model_meta, max_depth=max_depth, cur_depth=cur_depth + 1
                )
            ),
        )


def walk_select_relateds(obj, getters):
    """Walk select related of obj from getters."""
    for getter, nested_getters in getters:
        related_obj = getter(obj)
        if related_obj is None:
            # We don't need to seal a None relation or any of its children.
            continue
        yield related_obj
        yield from walk_select_relateds(related_obj, nested_getters)


class SealedModelIterable(models.query.ModelIterable):
    def _sealed_iterator(self):
        """Iterate over objects and seal them."""
        objs = super().__iter__()
        for obj in objs:
            obj._state.sealed = True
            yield obj

    def _sealed_related_iterator(self, related_walker):
        """Iterate over objects and seal them and their select related."""
        for obj in self._sealed_iterator():
            for related_obj in related_walker(obj):
                related_obj._state.sealed = True
            yield obj

    def __iter__(self):
        query = self.queryset.query
        select_related = query.select_related
        if select_related:
            opts = self.queryset.model._meta
            if isinstance(select_related, dict):
                select_related_getters = tuple(
                    get_restricted_select_related_getters(
                        self.queryset.query.select_related, opts
                    )
                )
            else:
                select_related_getters = tuple(
                    get_unrestricted_select_related_getters(
                        opts, max_depth=query.max_depth
                    )
                )
            related_walker = partial(
                walk_select_relateds, getters=select_related_getters
            )
            iterator = self._sealed_related_iterator(related_walker)
        else:
            iterator = self._sealed_iterator()
        yield from iterator


class SealableQuerySet(models.QuerySet):
    _base_manager_class = None

    def as_manager(cls, seal=None):
        manager = cls._base_manager_class.from_queryset(cls)(seal=seal)
        manager._built_with_as_manager = True
        return manager

    as_manager.queryset_only = True
    as_manager = classmethod(as_manager)

    def seal(self, iterable_class=SealedModelIterable):
        if self._fields is not None:
            raise TypeError("Cannot call seal() after .values() or .values_list()")
        if not issubclass(iterable_class, SealedModelIterable):
            raise TypeError(
                "iterable_class %r is not a subclass of SealedModelIterable"
                % iterable_class
            )
        clone = self._clone()
        clone._iterable_class = iterable_class
        return clone



from django.apps import apps
from django.core import checks
from django.db import models
from django.db.models.fields.related import lazy_related_operation

from . import descriptors
from .query import SealableQuerySet


class BaseSealableManager(models.manager.Manager):
    def __init__(self, seal=None):
        self._seal_queryset = seal
        super().__init__()

    def _get_model(self):
        return self._model

    def _set_model(self, model):
        self._model = model
        if self._seal_queryset is None:
            self._seal_queryset = getattr(model, "_seal_managers", None)

    # Intercept .model assignment to inherit ._seal_managers as the
    # contribute_to_class() method is not called abstract model inheritance
    # of managers.
    model = property(_get_model, _set_model)

    def get_queryset(self):
        queryset = super().get_queryset()
        if self._seal_queryset:
            queryset = queryset.seal()
        return queryset

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if not issubclass(self.model, SealableModel):
            if getattr(self, "_built_with_as_manager", False):
                origin = "%s.as_manager()" % self._queryset_class.__name__
            else:
                origin = self.__class__.__name__
            errors.append(
                checks.Error(
                    "%s can only be used on seal.SealableModel subclasses." % origin,
                    id="seal.E001",
                    hint="Make %s inherit from seal.SealableModel."
                    % self.model._meta.label,
                    obj=self,
                )
            )
        return errors


SealableQuerySet._base_manager_class = BaseSealableManager
SealableManager = BaseSealableManager.from_queryset(SealableQuerySet, "SealableManager")


class SealableModel(models.Model):
    """
    Abstract model class that turns deferred and related fields accesses that
    would incur a database query into exceptions once sealed.
    """

    def __init_subclass__(cls, seal=None, **kwargs):
        if seal is None:
            seal = getattr(cls, "_seal_managers", seal)
        cls._seal_managers = seal
        return super().__init_subclass__(**kwargs)

    objects = SealableManager()

    class Meta:
        abstract = True

    def seal(self):
        """
        Seal the instance to turn deferred and related fields access that would
        required fetching from the database into exceptions.
        """
        self._state.sealed = True

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        if not apps.is_installed("seal"):
            errors.append(
                checks.Error(
                    "The 'seal' app must be installed to use `SealableModel`.",
                    id="seal.E002",
                    hint="Add 'seal' to your `INSTALLED_APPS`.",
                    obj=cls,
                )
            )
        return errors


def make_descriptor_sealable(model, attname):
    """
    Make a descriptor sealable if a sealable class is defined.
    """
    try:
        descriptor = getattr(model, attname)
    except AttributeError:
        # Handle hidden reverse accessor case. e.g. related_name='+'
        return
    sealable_descriptor_class = descriptors.sealable_descriptor_classes.get(
        descriptor.__class__
    )
    if sealable_descriptor_class:
        descriptor.__class__ = sealable_descriptor_class


def make_remote_field_descriptor_sealable(model, related_model, remote_field):
    """
    Make a remote field descriptor sealable if a sealable class is defined.
    """
    if not issubclass(related_model, SealableModel):
        return
    accessor_name = remote_field.get_accessor_name()
    # Self-referential many-to-many fields don't have a reverse accessor.
    if accessor_name is None:
        return
    make_descriptor_sealable(related_model, accessor_name)


def make_model_sealable(model):
    """
    Replace forward fields descriptors by sealable ones and reverse fields
    descriptors attached to SealableModel subclasses as well.

    This function should be called on a third-party model once all apps are
    done loading models such as from an AppConfig.ready().
    """
    opts = model._meta
    for field in opts.local_fields + opts.local_many_to_many + opts.private_fields:
        name = field.name
        attnames = {name, getattr(field, "attname", name)}
        for attname in attnames:
            make_descriptor_sealable(model, attname)
        remote_field = field.remote_field
        if remote_field:
            # Use lazy_related_operation because lazy relationships might not
            # be resolved yet.
            lazy_related_operation(
                make_remote_field_descriptor_sealable,
                model,
                remote_field.model,
                remote_field=remote_field,
            )
    # Non SealableModel subclasses won't have remote fields descriptors
    # attached to them made sealable so make sure to make locally defined
    # related objects sealable.
    if not issubclass(model, SealableModel):
        for related_object in opts.related_objects:
            make_descriptor_sealable(model, related_object.get_accessor_name())



import pickle
import warnings

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Prefetch
from django.db.models.query import ModelIterable
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from seal.descriptors import _SealedRelatedQuerySet
from seal.exceptions import UnsealedAttributeAccess
from seal.models import make_model_sealable
from seal.query import SealableQuerySet, SealedModelIterable

from .models import (
    Climate,
    GreatSeaLion,
    Island,
    Leak,
    Location,
    Nickname,
    SeaGull,
    SeaLion,
)


class SealableQuerySetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.location = Location.objects.create(latitude=51.585474, longitude=156.634331)
        cls.climate = Climate.objects.create(temperature=100)
        cls.location.climates.add(cls.climate)
        cls.leak = Leak.objects.create(description="Salt water")
        cls.great_sealion = GreatSeaLion.objects.create(
            height=1,
            weight=100,
            location=cls.location,
            leak=cls.leak,
            leak_o2o=cls.leak,
        )
        cls.sealion = cls.great_sealion.sealion_ptr
        cls.sealion.previous_locations.add(cls.location)
        cls.gull = SeaGull.objects.create(sealion=cls.sealion)
        cls.nickname = Nickname.objects.create(
            name="Jonathan Livingston", content_object=cls.gull
        )
        cls.island = Island.objects.create(location=cls.location)
        tests_models = tuple(apps.get_app_config("tests").get_models())
        ContentType.objects.get_for_models(*tests_models, for_concrete_models=True)

    def setUp(self):
        warnings.filterwarnings("error", category=UnsealedAttributeAccess)
        self.addCleanup(warnings.resetwarnings)

    def test_state_sealed_assigned(self):
        instance = SeaLion.objects.seal().get()
        self.assertTrue(instance._state.sealed)

    def test_sealed_deferred_field(self):
        instance = SeaLion.objects.seal().defer("weight").get()
        message = (
            'Attempt to fetch deferred field "weight" on sealed <SeaLion instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.weight
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_deferred_field(self):
        instance = SeaLion.objects.defer("weight").get()
        self.assertEqual(instance.weight, 100)

    def test_sealed_foreign_key(self):
        instance = SeaLion.objects.seal().get()
        message = (
            'Attempt to fetch related field "location" on sealed <SeaLion instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.location
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_foreign_key(self):
        instance = SeaLion.objects.get()
        self.assertEqual(instance.location, self.location)

    def test_sealed_select_related_foreign_key(self):
        instance = SeaLion.objects.select_related("location").seal().get()
        self.assertEqual(instance.location, self.location)
        instance = SeaGull.objects.select_related("sealion").seal().get()
        message = (
            'Attempt to fetch related field "location" on sealed <SeaLion instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.sealion.location
        instance = SeaGull.objects.select_related("sealion__location").seal().get()
        self.assertEqual(instance.sealion.location, self.location)
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_select_related_none_foreign_key(self):
        SeaLion.objects.update(location=None)
        instance = SeaLion.objects.select_related("location").seal().get()
        self.assertIsNone(instance.location)
        SeaGull.objects.update(sealion=None)
        instance = SeaGull.objects.select_related("sealion__location").seal().get()
        self.assertIsNone(instance.sealion)

    def test_select_related_foreign_key_leak(self):
        instance = SeaLion.objects.get()
        self.assertEqual(instance.leak.description, self.leak.description)

        instance = SeaLion.objects.select_related("leak").get()
        self.assertEqual(instance.leak.description, self.leak.description)

    def test_select_related_foreign_key_leak_o2o(self):
        instance = SeaLion.objects.get()
        self.assertEqual(instance.leak_o2o.description, self.leak.description)

        instance = SeaLion.objects.select_related("leak_o2o").get()
        self.assertEqual(instance.leak_o2o.description, self.leak.description)

    def test_sealed_select_related_foreign_key_leak(self):
        instance = (
            SeaLion.objects.select_related("leak")
            .defer("leak__description")
            .seal()
            .get()
        )
        with self.assertNumQueries(1):
            self.assertEqual(instance.leak.description, self.leak.description)

    def test_sealed_select_related_foreign_key_leak_o2o(self):
        instance = (
            SeaLion.objects.select_related("leak_o2o")
            .defer("leak_o2o__description")
            .seal()
            .get()
        )
        with self.assertNumQueries(1):
            self.assertEqual(instance.leak_o2o.description, self.leak.description)

    def test_sealed_select_related_deferred_field(self):
        instance = (
            SeaGull.objects.select_related(
                "sealion__location",
            )
            .only("sealion__location__latitude")
            .seal()
            .get()
        )
        self.assertEqual(instance.sealion.location, self.location)
        self.assertEqual(instance.sealion.location.latitude, self.location.latitude)
        message = (
            'Attempt to fetch deferred field "longitude" on sealed <Location instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.sealion.location.longitude
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_one_to_one(self):
        instance = SeaGull.objects.seal().get()
        message = (
            'Attempt to fetch related field "sealion" on sealed <SeaGull instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.sealion
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_one_to_one(self):
        instance = SeaGull.objects.get()
        self.assertEqual(instance.sealion, self.sealion)

    def test_sealed_select_related_one_to_one(self):
        instance = SeaGull.objects.select_related("sealion").seal().get()
        self.assertEqual(instance.sealion, self.sealion)

    def test_sealed_select_related_reverse_one_to_one(self):
        instance = SeaLion.objects.select_related("gull").seal().get()
        self.assertEqual(instance.gull, self.gull)
        self.gull.sealion = None
        self.gull.save(update_fields={"sealion"})
        instance = SeaLion.objects.select_related("gull").seal().get()
        with self.assertRaises(SeaLion.gull.RelatedObjectDoesNotExist):
            instance.gull

    def test_sealed_select_related_unrestricted(self):
        instance = Island.objects.select_related().seal().get()
        self.assertEqual(instance.location, self.location)
        instance = SeaLion.objects.select_related().seal().get()
        message = (
            'Attempt to fetch related field "location" on sealed <SeaLion instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            # null=True relationships are not followed when using
            # an unrestricted select_related()
            instance.location
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_prefetch_related_reverse_one_to_one(self):
        instance = SeaLion.objects.prefetch_related("gull").seal().get()
        self.assertEqual(instance.gull, self.gull)
        self.gull.sealion = None
        self.gull.save(update_fields={"sealion"})
        instance = SeaLion.objects.prefetch_related("gull").seal().get()
        with self.assertRaises(SeaLion.gull.RelatedObjectDoesNotExist):
            instance.gull

    def test_sealed_many_to_many(self):
        instance = SeaLion.objects.seal().get()
        message = 'Attempt to fetch many-to-many field "previous_locations" on sealed <SeaLion instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_locations.all())
        self.assertEqual(ctx.filename, __file__)
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.previous_locations.all()[0]
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_many_to_many_queryset(self):
        instance = SeaLion.objects.seal().get()
        self.assertEqual(
            instance.previous_locations.get(pk=self.location.pk), self.location
        )
        self.assertFalse(
            isinstance(
                instance.previous_locations.filter(pk=self.location.pk),
                _SealedRelatedQuerySet,
            )
        )

    def test_not_sealed_many_to_many(self):
        instance = SeaLion.objects.get()
        self.assertSequenceEqual(instance.previous_locations.all(), [self.location])

    def test_sealed_string_prefetched_many_to_many(self):
        instance = SeaLion.objects.prefetch_related("previous_locations").seal().get()
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
        instance = instance.previous_locations.all()[0]
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_visitors.all())
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_prefetch_prefetched_many_to_many(self):
        instance = (
            SeaLion.objects.prefetch_related(
                Prefetch("previous_locations"),
            )
            .seal()
            .get()
        )
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
        instance = instance.previous_locations.all()[0]
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_visitors.all())
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_prefetch_queryset_prefetched_many_to_many(self):
        instance = (
            SeaLion.objects.prefetch_related(
                Prefetch("previous_locations", Location.objects.all()),
            )
            .seal()
            .get()
        )
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
        instance = instance.previous_locations.all()[0]
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_visitors.all())
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_string_prefetched_nested_many_to_many(self):
        with self.assertNumQueries(3):
            instance = (
                SeaLion.objects.prefetch_related(
                    "previous_locations__previous_visitors"
                )
                .seal()
                .get()
            )
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
            self.assertSequenceEqual(
                instance.previous_locations.all()[0].previous_visitors.all(),
                [self.sealion],
            )
        instance = instance.previous_locations.all()[0].previous_visitors.all()[0]
        message = 'Attempt to fetch many-to-many field "previous_locations" on sealed <SeaLion instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_locations.all())
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_prefetch_prefetched_nested_many_to_many(self):
        instance = (
            SeaLion.objects.prefetch_related(
                Prefetch("previous_locations__previous_visitors"),
            )
            .seal()
            .get()
        )
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
            self.assertSequenceEqual(
                instance.previous_locations.all()[0].previous_visitors.all(),
                [self.sealion],
            )
        instance = instance.previous_locations.all()[0].previous_visitors.all()[0]
        message = 'Attempt to fetch many-to-many field "previous_locations" on sealed <SeaLion instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_locations.all())
        self.assertEqual(ctx.filename, __file__)

    def test_prefetched_sealed_many_to_many(self):
        instance = SeaLion.objects.prefetch_related(
            Prefetch("previous_locations", Location.objects.seal()),
        ).get()
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.previous_locations.all(), [self.location])
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_locations.all()[0].previous_visitors.all())
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_deferred_parent_link(self):
        instance = GreatSeaLion.objects.only("pk").seal().get()
        message = 'Attempt to fetch related field "sealion_ptr" on sealed <GreatSeaLion instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.sealion_ptr
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_parent_link(self):
        instance = GreatSeaLion.objects.only("pk").get()
        self.assertEqual(instance.sealion_ptr, self.sealion)

    def test_sealed_parent_link(self):
        instance = GreatSeaLion.objects.seal().get()
        with self.assertNumQueries(0):
            self.assertEqual(instance.sealion_ptr, self.sealion)

    def test_sealed_generic_foreign_key(self):
        instance = Nickname.objects.seal().get()
        message = 'Attempt to fetch related field "content_object" on sealed <Nickname instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.content_object
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_generic_foreign_key(self):
        instance = Nickname.objects.get()
        self.assertEqual(instance.content_object, self.gull)

    def test_sealed_prefetch_related_generic_foreign_key(self):
        instance = Nickname.objects.prefetch_related("content_object").seal().get()
        with self.assertNumQueries(0):
            self.assertEqual(instance.content_object, self.gull)

    def test_sealed_reverse_foreign_key(self):
        instance = Location.objects.seal().get()
        message = 'Attempt to fetch many-to-many field "visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.visitors.all())
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_reverse_foreign_key(self):
        instance = Location.objects.get()
        self.assertSequenceEqual(instance.visitors.all(), [self.sealion])

    def test_sealed_prefetched_reverse_foreign_key(self):
        instance = Location.objects.prefetch_related("visitors").seal().get()
        self.assertSequenceEqual(instance.visitors.all(), [self.sealion])

    def test_sealed_reverse_parent_link(self):
        instance = SeaLion.objects.seal().get()
        message = (
            'Attempt to fetch related field "greatsealion" on sealed <SeaLion instance>'
        )
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.greatsealion
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_reverse_parent_link(self):
        instance = SeaLion.objects.get()
        self.assertEqual(instance.greatsealion, self.great_sealion)

    def test_sealed_select_related_reverse_parent_link(self):
        instance = SeaLion.objects.select_related("greatsealion").seal().get()
        self.assertEqual(instance.greatsealion, self.great_sealion)

    def test_sealed_reverse_many_to_many(self):
        instance = Location.objects.seal().get()
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.previous_visitors.all())
        self.assertEqual(ctx.filename, __file__)
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.previous_visitors.all()[0]
        self.assertEqual(ctx.filename, __file__)

    def test_sealed_reverse_many_to_many_queryset(self):
        instance = Location.objects.seal().get()
        self.assertEqual(
            instance.previous_visitors.get(pk=self.sealion.pk), self.sealion
        )
        self.assertFalse(
            isinstance(
                instance.previous_visitors.filter(pk=self.sealion.pk),
                _SealedRelatedQuerySet,
            )
        )

    def test_not_reverse_sealed_many_to_many(self):
        instance = Location.objects.get()
        self.assertSequenceEqual(instance.previous_visitors.all(), [self.sealion])

    def test_sealed_prefetched_reverse_many_to_many(self):
        instance = Location.objects.prefetch_related("previous_visitors").seal().get()
        self.assertSequenceEqual(instance.previous_visitors.all(), [self.sealion])

    def test_sealed_generic_relation(self):
        instance = SeaGull.objects.seal().get()
        message = 'Attempt to fetch many-to-many field "nicknames" on sealed <SeaGull instance>'
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            list(instance.nicknames.all())
        self.assertEqual(ctx.filename, __file__)
        with self.assertWarnsMessage(UnsealedAttributeAccess, message) as ctx:
            instance.nicknames.all()[0]
        self.assertEqual(ctx.filename, __file__)

    def test_not_sealed_generic_relation(self):
        instance = SeaGull.objects.get()
        self.assertSequenceEqual(instance.nicknames.all(), [self.nickname])

    def test_sealed_prefetched_generic_relation(self):
        instance = SeaGull.objects.prefetch_related("nicknames").seal().get()
        self.assertSequenceEqual(instance.nicknames.all(), [self.nickname])

    def test_sealed_prefetched_select_related_many_to_many(self):
        with self.assertNumQueries(2):
            instance = (
                SeaLion.objects.select_related(
                    "location",
                )
                .prefetch_related(
                    "location__climates",
                )
                .seal()
                .get()
            )
        with self.assertNumQueries(0):
            self.assertSequenceEqual(instance.location.climates.all(), [self.climate])

    def test_prefetch_without_related_name(self):
        location = Location.objects.prefetch_related("island_set").seal().get()
        self.assertSequenceEqual(location.island_set.all(), [self.island])

    def test_prefetch_combine(self):
        with self.assertNumQueries(6):
            instance = (
                SeaLion.objects.prefetch_related(
                    "location",
                    "location__climates",
                    "location__previous_visitors__location",
                    "location__previous_visitors__previous_locations",
                )
                .seal()
                .get()
            )
        with self.assertNumQueries(0):
            self.assertEqual(instance.location, self.location)
            self.assertSequenceEqual(instance.location.climates.all(), [self.climate])
            self.assertSequenceEqual(
                instance.location.previous_visitors.all(), [self.sealion]
            )
            self.assertEqual(
                instance.location.previous_visitors.all()[0].location, self.location
            )
            self.assertSequenceEqual(
                instance.location.previous_visitors.all()[0].previous_locations.all(),
                [self.location],
            )

    def test_sealed_prefetch_related_results_cache(self):
        """Some related managers fetch objects in get_prefetch_queryset()."""
        location_relations = ["climates", "related_locations", "visitors"]
        for relation in location_relations:
            with self.subTest(relation=relation), self.assertNumQueries(2):
                list(Location.objects.seal().prefetch_related(relation).all())
        sea_lion_relations = ["location", "previous_locations", "leak", "leak_o2o"]
        for relation in sea_lion_relations:
            with self.subTest(relation=relation), self.assertNumQueries(2):
                list(SeaLion.objects.seal().prefetch_related(relation).all())

    def test_sealed_prefetch_many_to_many_results(self):
        other_location = Location.objects.create(
            latitude=51.585474, longitude=156.634331
        )
        other_climate = Climate.objects.create(temperature=60)
        other_location.climates.add(other_climate)
        with self.assertNumQueries(2):
            results = list(
                Location.objects.seal().prefetch_related("climates").order_by("pk")
            )
        self.assertEqual(results, [self.location, other_location])
        with self.assertNumQueries(0):
            self.assertEqual(list(results[0].climates.all()), [self.climate])
            self.assertEqual(list(results[1].climates.all()), [other_climate])

    def test_sealed_prefetch_reverse_many_to_one_results(self):
        other_location = Location.objects.create(
            latitude=51.585474, longitude=156.634331
        )
        other_sealion = SeaLion.objects.create(
            height=1, weight=2, location=other_location
        )
        with self.assertNumQueries(2):
            results = list(
                Location.objects.seal().prefetch_related("visitors").order_by("pk")
            )
        self.assertEqual(results, [self.location, other_location])
        with self.assertNumQueries(0):
            self.assertEqual(list(results[0].visitors.all()), [self.sealion])
            self.assertEqual(list(results[1].visitors.all()), [other_sealion])

    def test_sealed_prefetch_reverse_generic_many_to_one_results(self):
        other_sealion = SeaLion.objects.create(height=1, weight=2)
        other_gull = SeaGull.objects.create(sealion=other_sealion)
        other_nickname = Nickname.objects.create(
            name="Test Nickname", content_object=other_gull
        )
        with self.assertNumQueries(2):
            results = list(
                SeaGull.objects.seal().prefetch_related("nicknames").order_by("pk")
            )
        self.assertEqual(results, [self.gull, other_gull])
        with self.assertNumQueries(0):
            self.assertEqual(list(results[0].nicknames.all()), [self.nickname])
            self.assertEqual(list(results[1].nicknames.all()), [other_nickname])

    def test_related_sealed_pickleability(self):
        location = Location.objects.prefetch_related("climates").seal().get()
        climates_dump = pickle.dumps(location.climates.all())
        climates = pickle.loads(climates_dump)
        with self.assertNumQueries(0):
            self.assertEqual(list(climates)[0], self.climate)


class SealableQuerySetInteractionTests(SimpleTestCase):
    def test_values_seal_disallowed(self):
        with self.assertRaisesMessage(
            TypeError, "Cannot call seal() after .values() or .values_list()"
        ):
            SeaGull.objects.values("id").seal()

    def test_values_list_seal_disallowed(self):
        with self.assertRaisesMessage(
            TypeError, "Cannot call seal() after .values() or .values_list()"
        ):
            SeaGull.objects.values_list("id").seal()

    def test_seal_sealable_model_iterable_subclass(self):
        class SealableModelIterableSubclass(SealedModelIterable):
            pass

        queryset = SeaGull.objects.seal(iterable_class=SealableModelIterableSubclass)
        self.assertIs(queryset._iterable_class, SealableModelIterableSubclass)

    def test_seal_non_sealable_model_iterable_subclass(self):
        message = (
            "iterable_class <class 'django.db.models.query.ModelIterable'> "
            "is not a subclass of SealedModelIterable"
        )
        with self.assertRaisesMessage(TypeError, message):
            SeaGull.objects.seal(iterable_class=ModelIterable)


class SealableQuerySetNonSealableModelTests(TestCase):
    """
    A SealableQuerySet should be usable on non SealableModel subclasses.
    """

    @classmethod
    def setUpTestData(cls):
        cls.location = Location.objects.create(latitude=51.585474, longitude=156.634331)
        cls.climate = Climate.objects.create(temperature=100)
        cls.location.climates.add(cls.climate)
        cls.sealion = SeaLion.objects.create(
            height=1, weight=100, location=cls.location
        )

    @isolate_apps("tests")
    def test_sealed_non_sealable_model(self):
        class NonSealableLocation(models.Model):
            class Meta:
                db_table = Location._meta.db_table

        queryset = SealableQuerySet(model=NonSealableLocation)
        instance = queryset.seal().get()
        self.assertTrue(instance._state.sealed)

    @isolate_apps("tests")
    def test_sealed_select_related_non_sealable_model(self):
        class NonSealableLocation(models.Model):
            class Meta:
                db_table = Location._meta.db_table

        class NonSealableSeaLion(models.Model):
            location = models.ForeignKey(NonSealableLocation, models.CASCADE)

            class Meta:
                db_table = SeaLion._meta.db_table

        queryset = SealableQuerySet(model=NonSealableSeaLion)
        instance = queryset.select_related("location").seal().get()
        self.assertTrue(instance._state.sealed)
        self.assertTrue(instance.location._state.sealed)

    @isolate_apps("tests")
    def test_sealed_prefetch_related_non_sealable_model(self):
        class NonSealableClimate(models.Model):
            objects = SealableQuerySet.as_manager()

            class Meta:
                db_table = Climate._meta.db_table

        class NonSealableLocationClimatesThrough(models.Model):
            climate = models.ForeignKey(NonSealableClimate, models.CASCADE)
            location = models.ForeignKey("NonSealableLocation", models.CASCADE)

            class Meta:
                db_table = Location.climates.through._meta.db_table

        class NonSealableLocation(models.Model):
            climates = models.ManyToManyField(
                NonSealableClimate, through=NonSealableLocationClimatesThrough
            )

            class Meta:
                db_table = Location._meta.db_table

        make_model_sealable(NonSealableLocation)
        queryset = SealableQuerySet(model=NonSealableLocation)
        instance = queryset.prefetch_related("climates").seal().get()
        self.assertTrue(instance._state.sealed)
        with self.assertNumQueries(0):
            self.assertTrue(instance.climates.all()[0]._state.sealed)



#!/usr/bin/env python
from setuptools import find_packages, setup

with open("README.rst") as file_:
    long_description = file_.read()

setup(
    name="django-seal",
    version="1.7.1",
    description=(
        "Allows ORM constructs to be sealed to prevent them from executing "
        "queries on attribute accesses."
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/charettes/django-seal",
    author="Simon Charette",
    author_email="simon.charette@zapier.com",
    install_requires=[
        "Django>=4.2",
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    license="MIT License",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Django :: 5.1",
        "Framework :: Django :: 5.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import warnings

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core import checks
from django.db import models
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from seal.descriptors import (
    SealableDeferredAttribute,
    SealableForwardManyToOneDescriptor,
    SealableForwardOneToOneDescriptor,
    SealableManyToManyDescriptor,
    SealableReverseManyToOneDescriptor,
    SealableReverseOneToOneDescriptor,
)
from seal.exceptions import UnsealedAttributeAccess
from seal.models import SealableManager, SealableModel, make_model_sealable
from seal.query import SealableQuerySet

from .models import GreatSeaLion, Location, Nickname, SeaGull, SeaLion


class SealableModelTests(SimpleTestCase):
    def setUp(self):
        warnings.filterwarnings("error", category=UnsealedAttributeAccess)
        self.addCleanup(warnings.resetwarnings)

    def test_installed_apps(self):
        with self.modify_settings(INSTALLED_APPS={"remove": ["seal"]}):
            self.assertEqual(
                SeaLion.check(),
                [
                    checks.Error(
                        "The 'seal' app must be installed to use `SealableModel`.",
                        id="seal.E002",
                        hint="Add 'seal' to your `INSTALLED_APPS`.",
                        obj=SeaLion,
                    )
                ],
            )

    def test_sealed_instance_deferred_attribute_access(self):
        instance = SeaLion.from_db("default", ["id"], [1])
        instance.seal()
        message = (
            'Attempt to fetch deferred field "weight" on sealed <SeaLion instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.weight

    def test_sealed_instance_deferred_foreign_key_attribute_access(self):
        instance = SeaLion.from_db("default", ["id"], [1])
        instance.seal()
        message = (
            'Attempt to fetch deferred field "location_id" on sealed <SeaLion instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.location_id

    def test_sealed_instance_foreign_key_access(self):
        instance = SeaLion.from_db("default", ["id", "location_id"], [1, 1])
        instance.seal()
        message = (
            'Attempt to fetch related field "location" on sealed <SeaLion instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.location

    def test_sealed_instance_reverse_foreign_key_access(self):
        instance = Location.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch many-to-many field "visitors" on sealed <Location instance>'
        visitors = instance.visitors.all()
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            list(visitors)

    def test_sealed_instance_one_to_one_access(self):
        instance = SeaGull.from_db("default", ["id", "sealion_id"], [1, 1])
        instance.seal()
        message = (
            'Attempt to fetch related field "sealion" on sealed <SeaGull instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.sealion

    def test_sealed_instance_reverse_one_to_one_access(self):
        instance = SeaLion.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch related field "gull" on sealed <SeaLion instance>'
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.gull

    def test_sealed_instance_parent_link_access(self):
        instance = SeaLion.from_db("default", ["id"], [1])
        instance.seal()
        message = (
            'Attempt to fetch related field "greatsealion" on sealed <SeaLion instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.greatsealion

    def test_sealed_instance_reverse_parent_link_access(self):
        instance = GreatSeaLion.from_db("default", ["sealion_ptr_id"], [1])
        instance.seal()
        message = 'Attempt to fetch related field "sealion_ptr" on sealed <GreatSeaLion instance>'
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.sealion_ptr

    def test_sealed_instance_reverse_parent_link_access_sealed(self):
        instance = GreatSeaLion.from_db(
            "default",
            [
                "id",
                "sealion_ptr_id",
                "height",
                "weight",
                "location_id",
                "leak_id",
                "leak_o2o_id",
            ],
            [1, 1, 1, 1, 1, 1, 1],
        )
        instance.seal()
        message = (
            'Attempt to fetch related field "location" on sealed <SeaLion instance>'
        )
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            instance.sealion_ptr.location

    def test_sealed_instance_m2m_access(self):
        instance = SeaLion.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch many-to-many field "previous_locations" on sealed <SeaLion instance>'
        previous_locations = instance.previous_locations.all()
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            list(previous_locations)

    def test_sealed_instance_reverse_m2m_access(self):
        instance = Location.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch many-to-many field "previous_visitors" on sealed <Location instance>'
        previous_visitors = instance.previous_visitors.all()
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            list(previous_visitors)

    def test_sealed_instance_self_referential_m2m_acccess(self):
        instance = Location.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch many-to-many field "related_locations" on sealed <Location instance>'
        previous_visitors = instance.related_locations.all()
        with self.assertRaisesMessage(UnsealedAttributeAccess, message):
            list(previous_visitors)


class ContentTypesSealableModelTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        tests_models = tuple(apps.get_app_config("tests").get_models())
        ContentType.objects.get_for_models(*tests_models, for_concrete_models=True)

    def setUp(self):
        warnings.filterwarnings("error", category=UnsealedAttributeAccess)
        self.addCleanup(warnings.resetwarnings)

    def test_sealed_instance_generic_foreign_key(self):
        instance = Nickname.from_db(
            "default", ["id", "content_type_id", "object_id"], [1, 1, 1]
        )
        instance.seal()
        message = 'Attempt to fetch related field "content_object" on sealed <Nickname instance>'
        with self.assertNumQueries(0), self.assertRaisesMessage(
            UnsealedAttributeAccess, message
        ):
            instance.content_object

    def test_sealed_instance_generic_relation(self):
        instance = SeaGull.from_db("default", ["id"], [1])
        instance.seal()
        message = 'Attempt to fetch many-to-many field "nicknames" on sealed <SeaGull instance>'
        nicknames = instance.nicknames.all()
        with self.assertNumQueries(0), self.assertRaisesMessage(
            UnsealedAttributeAccess, message
        ):
            list(nicknames)


class SealableManagerTests(SimpleTestCase):
    def test_isinstance_manager(self):
        """Manager classes are subclasses of Manager as many third-party apps expect."""
        self.assertIsInstance(SealableManager(), models.Manager)
        self.assertIsInstance(SealableQuerySet.as_manager(), models.Manager)

    @isolate_apps("tests")
    def test_declarative_seal(self):
        class SealedManagers(SealableModel):
            manager = SealableManager(seal=True)
            as_manager = SealableQuerySet.as_manager(seal=True)

        sealed_iterable_class = SealedManagers.objects.seal()._iterable_class
        self.assertIsNot(
            SealedManagers.objects.all()._iterable_class, sealed_iterable_class
        )
        self.assertIs(
            SealedManagers.manager.all()._iterable_class, sealed_iterable_class
        )
        self.assertIs(
            SealedManagers.as_manager.all()._iterable_class, sealed_iterable_class
        )

        class MixinInitSubclass:
            def __init_subclass__(cls, foo=None, **kwargs):
                cls.foo = foo
                super().__init_subclass__(**kwargs)

        class SealedBaseModel(MixinInitSubclass, SealableModel, foo="bar", seal=True):
            class Meta:
                abstract = True

        self.assertEqual(SealedBaseModel.foo, "bar")

        class SealedModel(SealedBaseModel):
            manager = SealableManager(seal=False)
            as_manager = SealableQuerySet.as_manager(seal=False)

        self.assertIs(SealedModel.objects.all()._iterable_class, sealed_iterable_class)
        self.assertIsNot(
            SealedModel.manager.all()._iterable_class, sealed_iterable_class
        )
        self.assertIsNot(
            SealedModel.as_manager.all()._iterable_class, sealed_iterable_class
        )

    @isolate_apps("tests")
    def test_non_sealable_model(self):
        class Foo(models.Model):
            manager = SealableManager()
            as_manager = SealableQuerySet.as_manager()

        self.assertEqual(
            Foo.manager.check(),
            [
                checks.Error(
                    "SealableManager can only be used on seal.SealableModel subclasses.",
                    id="seal.E001",
                    hint="Make tests.Foo inherit from seal.SealableModel.",
                    obj=Foo.manager,
                )
            ],
        )
        self.assertEqual(
            Foo.as_manager.check(),
            [
                checks.Error(
                    "SealableQuerySet.as_manager() can only be used on seal.SealableModel subclasses.",
                    id="seal.E001",
                    hint="Make tests.Foo inherit from seal.SealableModel.",
                    obj=Foo.as_manager,
                )
            ],
        )


class MakeModelSealableTests(SimpleTestCase):
    @isolate_apps("tests")
    def test_make_non_sealable_model_subclass(self):
        class Foo(models.Model):
            pass

        class Bar(models.Model):
            foo = models.BooleanField(default=False)
            fk = models.ForeignKey(Foo, models.CASCADE, related_name="fk_bar")
            o2o = models.OneToOneField(Foo, models.CASCADE, related_name="o2o_bar")
            m2m = models.ManyToManyField(Foo, related_name="m2m_bar")

        make_model_sealable(Bar)

        # Forward fields descriptors should have been made sealable.
        self.assertIsInstance(Bar.foo, SealableDeferredAttribute)
        self.assertIsInstance(Bar.fk, SealableForwardManyToOneDescriptor)
        self.assertIsInstance(Bar.o2o, SealableForwardOneToOneDescriptor)
        self.assertIsInstance(Bar.m2m, SealableManyToManyDescriptor)

        # But not the remote fields.
        self.assertNotIsInstance(Foo.fk_bar, SealableReverseManyToOneDescriptor)
        self.assertNotIsInstance(Foo.o2o_bar, SealableReverseOneToOneDescriptor)
        self.assertNotIsInstance(Foo.m2m_bar, SealableManyToManyDescriptor)

        # Should seal local related objects.
        make_model_sealable(Foo)
        self.assertIsInstance(Foo.fk_bar, SealableReverseManyToOneDescriptor)
        self.assertIsInstance(Foo.o2o_bar, SealableReverseOneToOneDescriptor)
        self.assertIsInstance(Foo.m2m_bar, SealableManyToManyDescriptor)



import warnings
from functools import lru_cache

from django.db.models import QuerySet
from django.db.models.fields import DeferredAttribute
from django.db.models.fields.related import (
    ForeignKeyDeferredAttribute,
    ForwardManyToOneDescriptor,
    ForwardOneToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.utils.functional import cached_property

from . import models
from .exceptions import UnsealedAttributeAccess
from .query import SealableQuerySet


def _bare_repr(instance):
    return "<%s instance>" % instance.__class__.__name__


class _SealedRelatedQuerySet(QuerySet):
    """
    QuerySet that prevents any fetching from taking place on its current form.

    As soon as the query is cloned it gets unsealed.
    """

    def _clone(self, *args, **kwargs):
        clone = super()._clone(*args, **kwargs)
        clone.__class__ = self._unsealed_class
        return clone

    def __getitem__(self, item):
        if self._result_cache is None:
            warnings.warn(
                self._sealed_warning, category=UnsealedAttributeAccess, stacklevel=2
            )
        return super().__getitem__(item)

    def _fetch_all(self):
        if self._result_cache is None:
            warnings.warn(
                self._sealed_warning, category=UnsealedAttributeAccess, stacklevel=3
            )
        super()._fetch_all()

    def __reduce__(self):
        return (
            _unpickle_sealed_related_queryset,
            (self._unsealed_class,),
            self.__getstate__(),
        )


class SealedPrefetchMixin(object):
    def _get_default_prefetch_queryset(self):
        return self.get_queryset()

    def get_prefetch_queryset(self, instances, queryset=None):
        if queryset is None:
            queryset = self._get_default_prefetch_queryset()
        if getattr(instances[0]._state, "sealed", False) and isinstance(
            queryset, SealableQuerySet
        ):
            queryset = queryset.seal()
        return super().get_prefetch_queryset(instances, queryset)

    def get_prefetch_querysets(self, instances, querysets=None):
        if querysets is None:
            querysets = [self._get_default_prefetch_queryset()]
        if getattr(instances[0]._state, "sealed", False):
            querysets = [
                queryset.seal() if isinstance(queryset, SealableQuerySet) else queryset
                for queryset in querysets
            ]
        return super().get_prefetch_querysets(instances, querysets)


@lru_cache(maxsize=100)
def _sealed_related_queryset_type_factory(queryset_cls):
    if issubclass(queryset_cls, _SealedRelatedQuerySet):
        return queryset_cls
    return type(
        f"Sealed{queryset_cls.__name__}",
        (_SealedRelatedQuerySet, queryset_cls),
        {
            "_unsealed_class": queryset_cls,
        },
    )


def _unpickle_sealed_related_queryset(queryset_cls):
    cls = _sealed_related_queryset_type_factory(queryset_cls)
    return cls.__new__(cls)


def seal_related_queryset(queryset, warning):
    """
    Seal a related queryset to prevent it from being fetched directly.
    """
    queryset.__class__ = _sealed_related_queryset_type_factory(queryset.__class__)
    queryset._sealed_warning = warning
    return queryset


def create_sealable_related_manager(related_manager_cls, field_name):
    class SealableRelatedManager(SealedPrefetchMixin, related_manager_cls):
        def _get_default_prefetch_queryset(self):
            # By-pass `related_manager_cls.get_queryset()` as that's the default
            # for dynamically created manager's `get_prefetch_queryset` when
            # no `queryset` is specified.
            return super(related_manager_cls, self).get_queryset()

        def get_queryset(self):
            if getattr(self.instance._state, "sealed", False):
                try:
                    prefetch_cache_name = self.prefetch_cache_name
                except AttributeError:
                    prefetch_cache_name = self.field.related_query_name()
                try:
                    return self.instance._prefetched_objects_cache[prefetch_cache_name]
                except (AttributeError, KeyError):
                    warning = (
                        'Attempt to fetch many-to-many field "%s" on sealed %s.'
                        % (
                            field_name,
                            _bare_repr(self.instance),
                        )
                    )
                    related_queryset = super().get_queryset()
                    return seal_related_queryset(related_queryset, warning)
            return super().get_queryset()

    return SealableRelatedManager


class SealableDeferredAttribute(DeferredAttribute):
    @cached_property
    def field_name(self):
        return self.field.attname

    def _check_parent_chain(self, instance, field_name=None):
        super()._check_parent_chain(instance)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        if (
            getattr(instance._state, "sealed", False)
            and instance.__dict__.get(self.field_name, self) is self
            and self._check_parent_chain(instance, self.field_name) is None
        ):
            message = 'Attempt to fetch deferred field "%s" on sealed %s.' % (
                self.field_name,
                _bare_repr(instance),
            )
            warnings.warn(message, category=UnsealedAttributeAccess, stacklevel=2)
        return super().__get__(instance, cls)


class SealableForwardOneToOneDescriptor(SealedPrefetchMixin, ForwardOneToOneDescriptor):
    def get_object(self, instance):
        sealed = getattr(instance._state, "sealed", False)
        if sealed:
            rel_model = self.field.remote_field.model
            if self.field.remote_field.parent_link and issubclass(
                rel_model, models.SealableModel
            ):
                deferred = instance.get_deferred_fields()
                # Because it's a parent link, all the data is available in the
                # instance, so populate the parent model with this data.

                fields = {field.attname for field in rel_model._meta.concrete_fields}

                # If any of the related model's fields are deferred, prevent
                # the query from being performed.
                if any(field in fields for field in deferred):
                    message = 'Attempt to fetch related field "%s" on sealed %s.' % (
                        self.field.name,
                        _bare_repr(instance),
                    )
                    warnings.warn(
                        message, category=UnsealedAttributeAccess, stacklevel=3
                    )
                else:
                    # When none of the fields inherited from the parent link
                    # are deferred ForwardOneToOneDescriptor.get_object() simply
                    # create an in-memory object from the existing field values.
                    # Make sure this in-memory instance is sealed as well.
                    obj = super().get_object(instance)
                    obj.seal()
                    return obj
            else:
                message = 'Attempt to fetch related field "%s" on sealed %s.' % (
                    self.field.name,
                    _bare_repr(instance),
                )
                warnings.warn(message, category=UnsealedAttributeAccess, stacklevel=3)
        return super().get_object(instance)


class SealableReverseOneToOneDescriptor(SealedPrefetchMixin, ReverseOneToOneDescriptor):
    def get_queryset(self, **hints):
        instance = hints.get("instance")
        if instance and getattr(instance._state, "sealed", False):
            message = 'Attempt to fetch related field "%s" on sealed %s.' % (
                self.related.name,
                _bare_repr(instance),
            )
            warnings.warn(message, category=UnsealedAttributeAccess, stacklevel=3)
        return super().get_queryset(**hints)


class SealableForwardManyToOneDescriptor(ForwardManyToOneDescriptor):
    def get_object(self, instance):
        if getattr(instance._state, "sealed", False):
            message = 'Attempt to fetch related field "%s" on sealed %s.' % (
                self.field.name,
                _bare_repr(instance),
            )
            warnings.warn(message, category=UnsealedAttributeAccess, stacklevel=3)
        return super().get_object(instance)


class SealableReverseManyToOneDescriptor(ReverseManyToOneDescriptor):
    @cached_property
    def related_manager_cls(self):
        related_manager_cls = super().related_manager_cls
        return create_sealable_related_manager(related_manager_cls, self.rel.name)


class SealableManyToManyDescriptor(ManyToManyDescriptor):
    @cached_property
    def related_manager_cls(self):
        related_manager_cls = super().related_manager_cls
        field_name = self.rel.name if self.reverse else self.field.name
        return create_sealable_related_manager(related_manager_cls, field_name)


class SealableForeignKeyDeferredAttribute(
    SealableDeferredAttribute, ForeignKeyDeferredAttribute
):
    pass


sealable_descriptor_classes = {
    DeferredAttribute: SealableDeferredAttribute,
    ForwardOneToOneDescriptor: SealableForwardOneToOneDescriptor,
    ReverseOneToOneDescriptor: SealableReverseOneToOneDescriptor,
    ForwardManyToOneDescriptor: SealableForwardManyToOneDescriptor,
    ReverseManyToOneDescriptor: SealableReverseManyToOneDescriptor,
    ManyToManyDescriptor: SealableManyToManyDescriptor,
    ForeignKeyDeferredAttribute: SealableForeignKeyDeferredAttribute,
}


def make_contenttypes_sealable():
    from django.contrib.contenttypes.fields import (
        GenericForeignKey,
        ReverseGenericManyToOneDescriptor,
    )

    # Ensure the function is idempotent.
    if GenericForeignKey in sealable_descriptor_classes:
        return

    class SealableGenericForeignKey(GenericForeignKey):
        def __get__(self, instance, cls=None):
            if instance is None:
                return self

            if getattr(instance._state, "sealed", False) and not self.is_cached(
                instance
            ):
                message = 'Attempt to fetch related field "%s" on sealed %s.' % (
                    self.name,
                    _bare_repr(instance),
                )
                warnings.warn(message, category=UnsealedAttributeAccess, stacklevel=2)

            return super().__get__(instance, cls=cls)

    class SealableReverseGenericManyToOneDescriptor(ReverseGenericManyToOneDescriptor):
        @cached_property
        def related_manager_cls(self):
            related_manager_cls = super().related_manager_cls
            return create_sealable_related_manager(related_manager_cls, self.field.name)

    sealable_descriptor_classes[GenericForeignKey] = SealableGenericForeignKey
    sealable_descriptor_classes[ReverseGenericManyToOneDescriptor] = (
        SealableReverseGenericManyToOneDescriptor
    )
